# Python imports
import signal
import subprocess
import threading
import json

# Lib imports
from gi.repository import GLib

# Application imports
from libs.dto.lsp_messages import LEN_HEADER, TYPE_HEADER, get_message_str, get_message_obj
from libs.dto.lsp_message_structs import \
    LSPResponseTypes, ClientRequest, ClientNotification, LSPResponseRequest, LSPResponseNotification
from .lsp_controller_events import LSPControllerEvents



def _log_list():
    def add_log_entry(title: str, message: any):
        ...

    def clear():
        ...



class LSPController(LSPControllerEvents):
    def __init__(self):
        super(LSPController).__init__()

        self._language      = "python3"

        # https://github.com/microsoft/multilspy/tree/main/src/multilspy/language_servers
        # initialize-params-slim.json was created off of jedi_language_server one
        self._init_params   = settings_manager.get_lsp_init_data()

        self._start_command = None
        self._lsp_pid       = -1
        self._message_id    = 0

        self.read_lock      = threading.Lock()
        self.write_lock     = threading.Lock()
        self.log_list       = _log_list
        self.request_list   = {}

        self._subscribe_to_events()


    def _subscribe_to_events(self):
        event_system.subscribe("textDocument/didOpen",    self._lsp_did_open)
        event_system.subscribe("textDocument/didSave",    self._lsp_did_save)
        event_system.subscribe("textDocument/didClose",   self._lsp_did_close)
        event_system.subscribe("textDocument/didChange",  self._lsp_did_change)
        event_system.subscribe("textDocument/definition", self._lsp_definition)
        event_system.subscribe("textDocument/completion", self._lsp_completion)


    def set_language(self, language):
        self._language = language

    def set_log_list(self, log_list):
        self.log_list = log_list

    def set_start_command(self, start_command: list):
        self._start_command = start_command

    def unset_start_command(self):
        self._start_command = None

    def send_notification(self, method: str, params: {} = {}):
        self._send_message( ClientNotification(method, params) )

    def send_request(self, method: str, params: {} = {}):
        self._send_message( ClientRequest(self._message_id, method, params) )

        self.request_list[self._message_id] = {"method": method}
        self._message_id += 1

    def _send_message(self, data: ClientRequest or ClientNotification):
        if not data or not hasattr(self, "lsp_process"): return

        message_str  = get_message_str(data)
        message_size = len(message_str)
        message      = f"Content-Length: {message_size}\r\n\r\n{message_str}"

        self.log_list.add_log_entry("Client", data)

        with self.write_lock:
            self.lsp_process.stdin.write( message.encode("utf-8") )
            self.lsp_process.stdin.flush()

    def get_message_id(self) -> int:
        return self._message_id

    def start_stop_lsp(self):
        if self._lsp_pid == -1:
            pid = self.start_lsp()
            if not pid: return

            self._lsp_pid = pid
            self._monitor_lsp_response()
        else:
            self.stop_lsp()
            self.log_list.clear()

    def start_lsp(self):
        if not self._start_command: return
        try:
            self.lsp_process = subprocess.Popen(
                self._start_command,
                stdout = subprocess.PIPE,
                stdin  = subprocess.PIPE
            )
        except Exception as e:
            self.log_list.add_log_entry(
                "LSP Client Error",
                LSPResponseRequest(
                    "2.0",
                    None,
                    {
                        "error": repr(e)
                    }
                )
            )

            return

        return self.lsp_process.pid

    def stop_lsp(self):
        if self._lsp_pid == -1: return

        self._lsp_pid    = -1
        self._message_id = 0
        self.lsp_process.terminate()

    # https://github.com/sr-lab/coqpyt/blob/master/coqpyt/lsp/json_rpc_endpoint.py#L65
    # Virtually this whole method unabashedly taken from ^ ...
    @daemon_threaded
    def _monitor_lsp_response(self):
        if not hasattr(self, "lsp_process"): return

        with self.read_lock:
            message_size = None
            while True:
                line = self.lsp_process.stdout.readline()
                if not line: return None  # Quit listener...

                line = line.decode("utf-8")
                if not line.endswith("\r\n"):
                    raise Exception(
                        "Bad header: missing newline"
                    )
                line = line[:-2]  # Strip the "\r\n"
                if line == "":    # We're done with the headers...
                    break
                elif line.startswith(LEN_HEADER):
                    line = line[len(LEN_HEADER) :]
                    if not line.isdigit():
                        raise Exception(
                            "Bad header: size is not int",
                        )
                    message_size = int(line)
                elif line.startswith(TYPE_HEADER):
                    # Not doing anything with type header, currently...
                    pass
                else:
                    line = line.split(LEN_HEADER)
                    if len(line) == 2:
                        message_size = line[1]
                    raise Exception(
                        "Bad header: unknown header"
                    )

            if not message_size: return

            _data        = self.lsp_process.stdout.read(message_size)
            data         = _data.decode("utf-8")
            message      = get_message_obj(data)
            keys         = message.keys()
            lsp_response = None

            if "result" in keys:
                lsp_response = LSPResponseRequest(**get_message_obj(data))

            if "method" in keys:
                lsp_response = LSPResponseNotification(**get_message_obj(data))

            response_id  = -1

            if not lsp_response: return
            GLib.idle_add(self.handle_lsp_response, lsp_response)

    def handle_lsp_response(self, lsp_response: LSPResponseTypes):
        self._monitor_lsp_response()
        self.log_list.add_log_entry("LSP Response", lsp_response)
        event_system.emit("respond-to-client", (get_message_str(lsp_response),))